@@ -4,16 +4,19 @@
   Author: Uwe Schulzweida
 
 */
-
-#include <algorithm>  // sort
+/*
+  cdo samplegridicon,icon_grid_0007_R02B06_G.nc,icon_grid_0019_R02B05_G.nc,icon_grid_0005_R02B04_G.nc \
+                     infileR02B06.nc outR02B04_mean.nc outR02B04_std.nc
+*/
 
 #include <cdi.h>
 
 #include "cdo_options.h"
+#include "cdo_omp.h"
 #include "process_int.h"
 #include "cdi_lockedIO.h"
 #include <mpim_grid.h>
-#include "grid_point_search.h"
+#include "grid_pointsearch.h"
 #include "verifygrid.h"
 #include "field_functions.h"
 
@@ -22,9 +25,8 @@ constexpr int MAX_CHILDS = 9;
 struct CellIndex
 {
   long ncells;
-  Varray<long> neighbor;  // neighbor cell index
-  Varray<long> parent;    // parent cell index
-  Varray<long> child;     // child cell index
+  Varray<long> parent;  // parent cell index
+  Varray<long> child;   // child cell index
   std::string filename;
 };
 
@@ -39,7 +41,7 @@ read_cellindex(const std::string &filename, CellIndex &cellindex)
 {
   auto streamID = stream_open_read_locked(filename.c_str());
   auto vlistID = streamInqVlist(streamID);
-  auto ngrids = vlistNgrids(vlistID);
+  auto ngrids = vlistNumGrids(vlistID);
   int gridID = -1;
   for (int index = 0; index < ngrids; ++index)
     {
@@ -91,15 +93,15 @@ read_cellindex(const std::string &filename, CellIndex &cellindex)
 
   for (long i = 0; i < ncells; ++i) cellindex.parent[i] = 0;
 
-  auto nrecs = streamInqTimestep(streamID, 0);
-  for (int recID = 0; recID < nrecs; ++recID)
+  auto numFields = streamInqTimestep(streamID, 0);
+  for (int fieldID = 0; fieldID < numFields; ++fieldID)
     {
       int varID, levelID;
-      size_t nmiss;
-      streamInqRecord(streamID, &varID, &levelID);
+      size_t numMissVals;
+      streamInqField(streamID, &varID, &levelID);
       if (varID == pid /* || varID == nid || varID == cid */)
         {
-          streamReadRecord(streamID, data.data(), &nmiss);
+          streamReadField(streamID, data.data(), &numMissVals);
           // if (varID == pid)
           {
             if (Options::cdoVerbose) cdo_print("Read parent_cell_index");
@@ -122,7 +124,7 @@ read_grid(const std::string &filename)
 {
   auto streamID = stream_open_read_locked(filename.c_str());
   auto vlistID = streamInqVlist(streamID);
-  auto ngrids = vlistNgrids(vlistID);
+  auto ngrids = vlistNumGrids(vlistID);
   int gridID = -1;
   for (int index = 0; index < ngrids; ++index)
     {
@@ -184,12 +186,6 @@ struct SortInfo
   int p, i;
 };
 
-static bool
-cmpsinfo(const SortInfo &a, const SortInfo &b)
-{
-  return a.p < b.p;
-}
-
 static void
 compute_child_from_parent(CellIndex &cellindex1, CellIndex &cellindex2)
 {
@@ -201,23 +197,25 @@ compute_child_from_parent(CellIndex &cellindex1, CellIndex &cellindex2)
   std::vector<long> idx1(ncells1);
   for (long i = 0; i < ncells1; ++i) idx1[i] = i;
   for (long i = 1; i < ncells1; ++i)
-    if (parent1[i] < parent1[i - 1])
-      {
-        if (Options::cdoVerbose) cdo_print("Sort parent index of %s!", cellindex1.filename);
-        std::vector<SortInfo> sinfo(ncells1);
-        for (long j = 0; j < ncells1; ++j)
-          {
-            sinfo[j].p = parent1[j];
-            sinfo[j].i = idx1[j];
-          }
-        std::sort(sinfo.begin(), sinfo.end(), cmpsinfo);
-        for (long j = 0; j < ncells1; ++j)
-          {
-            parent1[j] = sinfo[j].p;
-            idx1[j] = sinfo[j].i;
-          }
-        break;
-      }
+    {
+      if (parent1[i] < parent1[i - 1])
+        {
+          if (Options::cdoVerbose) cdo_print("Sort parent index of %s!", cellindex1.filename);
+          std::vector<SortInfo> sinfo(ncells1);
+          for (long j = 0; j < ncells1; ++j)
+            {
+              sinfo[j].p = parent1[j];
+              sinfo[j].i = idx1[j];
+            }
+          ranges::sort(sinfo, {}, &SortInfo::p);
+          for (long j = 0; j < ncells1; ++j)
+            {
+              parent1[j] = sinfo[j].p;
+              idx1[j] = sinfo[j].i;
+            }
+          break;
+        }
+    }
 
   auto ncells2 = cellindex2.ncells;
   cellindex2.child.resize(MAX_CHILDS * ncells2);
@@ -225,11 +223,11 @@ compute_child_from_parent(CellIndex &cellindex1, CellIndex &cellindex2)
   for (long i = 0; i < ncells2; ++i)
     {
       for (long k = 0; k < MAX_CHILDS; ++k) child2[i * MAX_CHILDS + k] = -1;
-      long j = find_index(i, ncells1, parent1);
+      auto j = find_index(i, ncells1, parent1);
       if (j < 0) continue;
       for (long k = 0; k < MAX_CHILDS; ++k)
         {
-          if (i != parent1[j + k]) break;
+          if ((size_t) (j + k) >= idx1.size() || i != parent1[j + k]) break;
           //  child2[i*MAX_CHILDS+k] = j+k;
           child2[i * MAX_CHILDS + k] = idx1[j + k];
         }
@@ -242,7 +240,7 @@ read_coordinates(const std::string &filename, long n, Varray<double> &lon, Varra
 {
   auto streamID = streamOpenRead(filename.c_str());
   auto vlistID = streamInqVlist(streamID);
-  auto ngrids = vlistNgrids(vlistID);
+  auto ngrids = vlistNumGrids(vlistID);
   int gridID = -1;
   for (int index = 0; index < ngrids; ++index)
     {
@@ -263,11 +261,12 @@ read_coordinates(const std::string &filename, long n, Varray<double> &lon, Varra
 }
 
 static void
-read_coordinates(const std::string &filename, long n, Varray<double> &lon, Varray<double> &lat, int nv, Varray<double> &lon_bnds, Varray<double> &lat_bnds)
+read_coordinates(const std::string &filename, long n, Varray<double> &lon, Varray<double> &lat, int nv, Varray<double> &lon_bnds,
+                 Varray<double> &lat_bnds)
 {
   auto streamID = streamOpenRead(filename.c_str());
   auto vlistID = streamInqVlist(streamID);
-  auto ngrids = vlistNgrids(vlistID);
+  auto ngrids = vlistNumGrids(vlistID);
   int gridID = -1;
   for (int index = 0; index < ngrids; ++index)
     {
@@ -296,35 +295,35 @@ read_coordinates(const std::string &filename, long n, Varray<double> &lon, Varra
   streamClose(streamID);
 }
 
-constexpr int MAX_SEARCH = 128;  // the triangles are distorted!
-
 static void
-compute_child_from_bounds(CellIndex &cellindex2, long ncells2, Varray<double> &grid_center_lon2, Varray<double> &grid_center_lat2,
-                          Varray<double> &grid_corner_lon2, Varray<double> &grid_corner_lat2, long ncells1, const Varray<double> &grid_center_lon1,
-                          const Varray<double> &grid_center_lat1)
+compute_child_from_bounds(CellIndex &cellindex2, Varray<double> &grid_center_lon2, Varray<double> &grid_center_lat2,
+                          Varray<double> &grid_corner_lon2, Varray<double> &grid_corner_lat2,
+                          const Varray<double> &grid_center_lon1, const Varray<double> &grid_center_lat1)
 {
   if (Options::cdoVerbose) cdo_print("%s", __func__);
 
-  bool xIsCyclic = false;
-  size_t dims[2];
-  dims[0] = ncells1;
-  dims[1] = 0;
-  GridPointSearch gps;
-  grid_point_search_create(gps, xIsCyclic, dims, ncells1, grid_center_lon1, grid_center_lat1);
-  knnWeightsType knnWeights(MAX_SEARCH);
-  size_t *nbrIndices = &knnWeights.m_indices[0];
+  long ncells2 = (long) grid_center_lon2.size();
+
+  GridPointsearch gps;
+  grid_pointsearch_create_unstruct(gps, grid_center_lon1, grid_center_lat1, true);
 
   int ncorner = 3;
-  Point3D centerPoint3D;
-  Varray<Point3D> cellCorners3D(4);
-  Varray<Point> cellCornersPlaneProjection(4);
+
+  constexpr int MaxSearch = 128;
+  std::vector<KnnData> knnDataMem;
+  for (int i = 0; i < Threading::ompNumThreads; ++i) knnDataMem.push_back(KnnData(MaxSearch));
 
   cellindex2.child.resize(MAX_CHILDS * ncells2);
   auto &child2 = cellindex2.child;
+
+#ifdef _OPENMP
+#pragma omp parallel for if (ncells2 > 20000) default(shared) schedule(static)
+#endif
   for (long cellNo2 = 0; cellNo2 < ncells2; ++cellNo2)
     {
       for (int k = 0; k < MAX_CHILDS; ++k) child2[cellNo2 * MAX_CHILDS + k] = -1;
 
+      Varray<Point3D> cellCorners3D(4);
       set_cell_corners_3D(ncorner, &grid_corner_lon2[cellNo2 * ncorner], &grid_corner_lat2[cellNo2 * ncorner], cellCorners3D);
       cellCorners3D[ncorner] = cellCorners3D[0];
 
@@ -334,6 +333,7 @@ compute_child_from_bounds(CellIndex &cellindex2, long ncells2, Varray<double> &g
           = (coordinateToIgnore == 1) ? cellCorners3D[0].X : ((coordinateToIgnore == 2) ? cellCorners3D[0].Y : cellCorners3D[0].Z);
       auto invertResult = (cval < 0.0);
 
+      Varray<Point> cellCornersPlaneProjection(4);
       set_cell_corners_plane_projection(coordinateToIgnore, ncorner, cellCorners3D, cellCornersPlaneProjection);
 
       auto isClockwise = are_polygon_vertices_arranged_in_clockwise_order(cellCornersPlaneProjection, ncorner + 1);
@@ -341,13 +341,17 @@ compute_child_from_bounds(CellIndex &cellindex2, long ncells2, Varray<double> &g
       if (invertResult) isClockwise = !isClockwise;
       if (isClockwise) continue;
 
-      grid_search_point(gps, grid_center_lon2[cellNo2], grid_center_lat2[cellNo2], knnWeights);
+      auto ompthID = cdo_omp_get_thread_num();
+      auto &knnData = knnDataMem[ompthID];
+
+      grid_search_point_unstruct(gps, PointLonLat{ grid_center_lon2[cellNo2], grid_center_lat2[cellNo2] }, knnData);
 
       int k = 0;
       double centerCoordinates[3];
-      for (int i = 0; i < MAX_SEARCH; ++i)
+      Point3D centerPoint3D;
+      for (int i = 0; i < MaxSearch; ++i)
         {
-          auto cellNo1 = nbrIndices[i];
+          auto cellNo1 = knnData.m_indices[i];
           if (cellNo1 < SIZE_MAX)
             {
               gcLLtoXYZ(grid_center_lon1[cellNo1], grid_center_lat1[cellNo1], centerCoordinates);
@@ -357,9 +361,8 @@ compute_child_from_bounds(CellIndex &cellindex2, long ncells2, Varray<double> &g
 
               auto centerPoint2D = set_center_point_plane_projection(coordinateToIgnore, centerPoint3D);
 
-              auto winding_number = winding_numbers_algorithm(cellCornersPlaneProjection, ncorner + 1, centerPoint2D);
-
-              if (winding_number != 0)
+              auto windingNumber = winding_numbers_algorithm(cellCornersPlaneProjection, ncorner + 1, centerPoint2D);
+              if (windingNumber != 0)
                 {
                   if (k >= MAX_CHILDS) cdo_abort("Internal problem, limit of MAX_CHILDS reached (limit=9).");
                   child2[cellNo2 * MAX_CHILDS + k++] = (long) cellNo1;
@@ -367,8 +370,6 @@ compute_child_from_bounds(CellIndex &cellindex2, long ncells2, Varray<double> &g
             }
         }
     }
-
-  grid_point_search_delete(gps);
 }
 
 static void
@@ -385,7 +386,7 @@ compute_child_from_coordinates(const CellIndex &cellindex1, CellIndex &cellindex
   read_coordinates(cellindex1.filename, ncells1, lon1, lat1);
   read_coordinates(cellindex2.filename, ncells2, lon2, lat2, 3, lon2_bnds, lat2_bnds);
 
-  compute_child_from_bounds(cellindex2, ncells2, lon2, lat2, lon2_bnds, lat2_bnds, ncells1, lon1, lat1);
+  compute_child_from_bounds(cellindex2, lon2, lat2, lon2_bnds, lat2_bnds, lon1, lat1);
 }
 
 static void
@@ -394,15 +395,13 @@ compute_child(CellIndex &cellindex1, CellIndex &cellindex2)
   bool lparent = true;
   auto ncells1 = cellindex1.ncells;
   const auto &parent1 = cellindex1.parent;
-  {
-    long i;
-    for (i = 0; i < ncells1; ++i)
-      if (parent1[i] >= 0) break;
-    if (i == ncells1) lparent = false;
-  }
 
-  if (lparent)
-    compute_child_from_parent(cellindex1, cellindex2);
+  long i;
+  for (i = 0; i < ncells1; ++i)
+    if (parent1[i] >= 0) break;
+  if (i == ncells1) lparent = false;
+
+  if (lparent) { compute_child_from_parent(cellindex1, cellindex2); }
   else
     {
       compute_child_from_coordinates(cellindex1, cellindex2);
@@ -466,8 +465,20 @@ samplegrid(double missval, long nci, Varray<CellIndex> &cellindex, const Varray<
     }
 }
 
-class ModuleSamplegridicon
+class Samplegridicon : public Process
 {
+public:
+  using Process::Process;
+  inline static CdoModule module = {
+    .name = "Samplegridicon",
+    .operators = { { "samplegridicon", 0, 0, "samplegrids" } },
+    .aliases = {},
+    .mode = EXPOSED,     // Module mode: 0:intern 1:extern
+    .number = CDI_REAL,  // Allowed number type
+    .constraints = { 1, 2, OnlyFirst },
+  };
+  inline static RegisterEntry<Samplegridicon> registration = RegisterEntry<Samplegridicon>(module);
+
   CdoStreamID streamID1;
   CdoStreamID streamID2;
   CdoStreamID streamID3;
@@ -491,12 +502,8 @@ class ModuleSamplegridicon
 
 public:
   void
-  init(void *process)
+  init() override
   {
-    cdo_initialize(process);
-
-    cdo_operator_add("samplegridicon", 0, 0, "sample grids");
-
     nsamplegrids = cdo_operator_argc();
     if (nsamplegrids < 2) cdo_abort("Parameter missing!");
 
@@ -532,7 +539,7 @@ public:
     vlistDefTaxis(vlistID2, taxisID2);
     vlistDefTaxis(vlistID3, taxisID3);
 
-    auto ngrids = vlistNgrids(vlistID1);
+    auto ngrids = vlistNumGrids(vlistID1);
     for (int index = 0; index < ngrids; ++index)
       {
         auto gridID = vlistGrid(vlistID1, index);
@@ -558,37 +565,36 @@ public:
   }
 
   void
-  run()
+  run() override
   {
     int tsID = 0;
     while (true)
       {
-        auto nrecs = cdo_stream_inq_timestep(streamID1, tsID);
-        if (nrecs == 0) break;
+        auto numFields = cdo_stream_inq_timestep(streamID1, tsID);
+        if (numFields == 0) break;
 
         cdo_taxis_copy_timestep(taxisID2, taxisID1);
         cdo_taxis_copy_timestep(taxisID3, taxisID1);
         cdo_def_timestep(streamID2, tsID);
         cdo_def_timestep(streamID3, tsID);
 
-        for (int recID = 0; recID < nrecs; ++recID)
+        for (int fieldID = 0; fieldID < numFields; ++fieldID)
           {
-            int varID, levelID;
-            cdo_inq_record(streamID1, &varID, &levelID);
-            size_t nmiss;
-            cdo_read_record(streamID1, array1.data(), &nmiss);
+            auto [varID, levelID] = cdo_inq_field(streamID1);
+            size_t numMissVals;
+            cdo_read_field(streamID1, array1.data(), &numMissVals);
 
             auto missval = vlistInqVarMissval(vlistID1, varID);
 
             samplegrid(missval, nsamplegrids, cellindex, array1, array2, array3);
 
-            nmiss = varray_num_mv(gridsize2, array2, missval);
-            cdo_def_record(streamID2, varID, levelID);
-            cdo_write_record(streamID2, array2.data(), nmiss);
+            numMissVals = varray_num_mv(gridsize2, array2, missval);
+            cdo_def_field(streamID2, varID, levelID);
+            cdo_write_field(streamID2, array2.data(), numMissVals);
 
-            nmiss = varray_num_mv(gridsize2, array3, missval);
-            cdo_def_record(streamID3, varID, levelID);
-            cdo_write_record(streamID3, array3.data(), nmiss);
+            numMissVals = varray_num_mv(gridsize2, array3, missval);
+            cdo_def_field(streamID3, varID, levelID);
+            cdo_write_field(streamID3, array3.data(), numMissVals);
           }
 
         tsID++;
@@ -596,26 +602,11 @@ public:
   }
 
   void
-  close()
+  close() override
   {
     cdo_stream_close(streamID3);
     cdo_stream_close(streamID2);
     cdo_stream_close(streamID1);
-
     vlistDestroy(vlistID2);
-    gridDestroy(gridID2);
-
-    cdo_finish();
   }
 };
-
-void *
-Samplegridicon(void *process)
-{
-  ModuleSamplegridicon samplegridicon;
-  samplegridicon.init(process);
-  samplegridicon.run();
-  samplegridicon.close();
-
-  return nullptr;
-}
