@@ -19,8 +19,19 @@
 #include "util_files.h"
 #include "util_string.h"
 
-class ModuleSplitsel
+class Splitsel : public Process
 {
+public:
+  using Process::Process;
+  inline static CdoModule module = {
+    .name = "Splitsel",
+    .operators = { { "splitsel", SplitselHelp } },
+    .aliases = {},
+    .mode = EXPOSED,     // Module mode: 0:intern 1:extern
+    .number = CDI_BOTH,  // Allowed number type
+    .constraints = { 1, OBASE, OnlyFirst },
+  };
+  inline static RegisterEntry<Splitsel> registration = RegisterEntry<Splitsel>(module);
 
   int noffset;
   CdoStreamID streamID1;
@@ -43,16 +54,11 @@ class ModuleSplitsel
 
 public:
   void
-  init(void *process)
+  init() override
   {
-    cdo_initialize(process);
-
-    if (process_self().m_ID != 0) cdo_abort("This operator can't be combined with other operators!");
 
     dataIsUnchanged = data_is_unchanged();
 
-    cdo_operator_add("splitsel", 0, 0, nullptr);
-
     // operator_input_arg("numSets <noffset <numSkip>>");
 
     auto nargc = cdo_operator_argc();
@@ -82,7 +88,7 @@ public:
     taxisID2 = taxisDuplicate(taxisID1);
     vlistDefTaxis(vlistID2, taxisID2);
 
-    varList_init(varList1, vlistID1);
+    varList1 = VarList(vlistID1);
 
     fileSuffix = FileUtils::gen_suffix(cdo_inq_filetype(streamID1), vlistID1, cdo_get_stream_name(0));
 
@@ -93,16 +99,15 @@ public:
       array.resize(gridsizemax);
     }
 
-    haveConstVars = (varList_numConstVars(varList1) > 0);
-
+    haveConstVars = (varList1.numConstVars() > 0);
     if (haveConstVars)
       {
-        int numVars = varList1.size();
+        int numVars = varList1.numVars();
         fields.resize(numVars);
 
         for (int varID = 0; varID < numVars; ++varID)
           {
-            const auto &var = varList1[varID];
+            const auto &var = varList1.vars[varID];
             if (var.isConstant)
               {
                 fields[varID].resize(var.nlevels);
@@ -117,33 +122,32 @@ public:
   }
 
   void
-  run()
+  run() override
   {
     int index = 1;
     int tsID;
-    int nrecs = 0;
+    int numFields = 0;
 
     // offset
     for (tsID = 0; tsID < noffset; ++tsID)
       {
-        nrecs = cdo_stream_inq_timestep(streamID1, tsID);
-        if (nrecs == 0)
+        numFields = cdo_stream_inq_timestep(streamID1, tsID);
+        if (numFields == 0)
           {
             cdo_warning("noffset is larger than number of timesteps!");
             return;
           }
 
         if (tsID == 0 && haveConstVars)
-          for (int recID = 0; recID < nrecs; ++recID)
+          for (int fieldID = 0; fieldID < numFields; ++fieldID)
             {
-              int varID, levelID;
-              cdo_inq_record(streamID1, &varID, &levelID);
+              auto [varID, levelID] = cdo_inq_field(streamID1);
 
-              const auto &var = varList1[varID];
+              const auto &var = varList1.vars[varID];
               if (var.isConstant)
                 {
                   auto &field = fields[varID][levelID];
-                  cdo_read_record(streamID1, field.vec_d.data(), &field.nmiss);
+                  cdo_read_field(streamID1, field.vec_d.data(), &field.numMissVals);
                 }
             }
       }
@@ -160,8 +164,8 @@ public:
 
         for (int i = 0; i < numSets; ++i)
           {
-            nrecs = cdo_stream_inq_timestep(streamID1, tsID);
-            if (nrecs == 0) break;
+            numFields = cdo_stream_inq_timestep(streamID1, tsID);
+            if (numFields == 0) break;
 
             cdo_taxis_copy_timestep(taxisID2, taxisID1);
 
@@ -175,43 +179,42 @@ public:
 
             if (tsID > 0 && tsID2 == 0 && haveConstVars)
               {
-                int numVars = varList1.size();
+                int numVars = varList1.numVars();
                 for (int varID = 0; varID < numVars; ++varID)
                   {
-                    const auto &var = varList1[varID];
+                    const auto &var = varList1.vars[varID];
                     if (var.isConstant)
                       {
                         for (int levelID = 0; levelID < var.nlevels; ++levelID)
                           {
-                            cdo_def_record(streamID2, varID, levelID);
+                            cdo_def_field(streamID2, varID, levelID);
                             auto &field = fields[varID][levelID];
-                            cdo_write_record(streamID2, field.vec_d.data(), field.nmiss);
+                            cdo_write_field(streamID2, field.vec_d.data(), field.numMissVals);
                           }
                       }
                   }
               }
 
-            for (int recID = 0; recID < nrecs; ++recID)
+            for (int fieldID = 0; fieldID < numFields; ++fieldID)
               {
-                int varID, levelID;
-                cdo_inq_record(streamID1, &varID, &levelID);
-                cdo_def_record(streamID2, varID, levelID);
+                auto [varID, levelID] = cdo_inq_field(streamID1);
+                cdo_def_field(streamID2, varID, levelID);
 
-                if (dataIsUnchanged && !(tsID == 0 && haveConstVars)) { cdo_copy_record(streamID2, streamID1); }
+                if (dataIsUnchanged && !(tsID == 0 && haveConstVars)) { cdo_copy_field(streamID2, streamID1); }
                 else
                   {
-                    size_t nmiss;
-                    cdo_read_record(streamID1, array.data(), &nmiss);
-                    cdo_write_record(streamID2, array.data(), nmiss);
+                    size_t numMissVals;
+                    cdo_read_field(streamID1, array.data(), &numMissVals);
+                    cdo_write_field(streamID2, array.data(), numMissVals);
 
                     if (tsID == 0 && haveConstVars)
                       {
-                        const auto &var = varList1[varID];
+                        const auto &var = varList1.vars[varID];
                         if (var.isConstant)
                           {
                             auto &field = fields[varID][levelID];
                             varray_copy(var.gridsize, array, field.vec_d);
-                            field.nmiss = nmiss;
+                            field.numMissVals = numMissVals;
                           }
                       }
                   }
@@ -222,7 +225,7 @@ public:
           }
 
         cdo_stream_close(streamID2);
-        if (nrecs == 0) break;
+        if (numFields == 0) break;
 
         if (cdo_stream_inq_timestep(streamID1, tsID) == 0) break;
 
@@ -239,23 +242,10 @@ public:
   }
 
   void
-  close()
+  close() override
   {
     cdo_stream_close(streamID1);
 
     vlistDestroy(vlistID2);
-
-    cdo_finish();
   }
 };
-
-void *
-Splitsel(void *process)
-{
-  ModuleSplitsel splitsel;
-  splitsel.init(process);
-  splitsel.run();
-  splitsel.close();
-
-  return nullptr;
-}
